import java.io.*;
import java.net.*;
import java.util.*;
import java.util.Map.Entry;

/**
 * UsaProxy - HTTP proxy for tracking and logging of user interactions on websites
 * <br><br>
 * Copyright (C) 2006  Monika Wnuk - Media Informatics Group at the University of Munich
 * <br><br>
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 * <br><br>
 * Class ClientRequest handles a single client request Request reading, forwarding to 
 * the web server/remote proxy and response processing are managed by this class. 
 * In Addition, log requests and file forwarding requests are routed to the instances 
 * of the respective classes, namely LogFile and ProxyScript.
 */
public class ClientRequest extends Thread {
	
	
	private UsaProxy			usaProxy;						  /**  associated parent proxy application */
	private SocketData			client,server;

    private String 				get;
    private String 				host;
   
    private boolean 			isHtmlStream;
    private boolean 			isNotFoundMessage;
    private StringBuffer 		data;
    
    private boolean				testSeiteRequested;
    private boolean 			proxyScriptRequested;
    private boolean 			logRequested;
    private String 				logData;						  /**  contains the data to log retrieved
    																	*  from the log request argument */
 
    private int					httpTrafficIndex;				  /**  index for txt-file suffix and as identifier
    																	*  in httpTraffic log entry */
    
    
    /** Constructor: create ClientRequest with a given client socket*/
    public ClientRequest (Socket socket, UsaProxy usaProxy) {

		this.client					= new SocketData(socket);
		this.usaProxy				= usaProxy;
		this.proxyScriptRequested 	= false;
		this.logRequested 			= false;
		this.testSeiteRequested 	= false;
	
		/** start thread */
		start();
    }
    
    /**
     *  Gets the input and output stream of the client and processes the request.
     *  <br><br>
     *  Depending on which action was requested either the proxyscript.js file is send,
     *  or a new log request processed, or a regular request forwarded to the server.
     *  <br><br>
     *  In case request was forwarded, the response is expected.
     */
    public void run () { 

		try { 
		    
		    /** Create client streams */
		    client.getInputStream();
		    client.getOutputStream();
		    
		    /** parse request and detect parameters */
		    get  	= "";
		    host 	= "";
		    client.setIn(processRequest(client.getIn()));
		    
		    
		    if (testSeiteRequested) {
		    	usaProxy.getProxyScript().send (client.getOut(), "testseite.html", client.getSocket(), "");
		    }
	
			/** if proxyscript.js is requested */
		    else if (proxyScriptRequested) {
		        
		        /** check if session ID available
		         *  if not: generate new random String */
		    	String sessionID = "";
		    	client.printHeaders();
		    	if(client.getHeaders().containsKey(HTTPData.HEADER_COOKIE))
		    		sessionID = client.getHeaders().get(HTTPData.HEADER_COOKIE);
		    	if (sessionID.indexOf(HTTPData.SESSION_ID_NAME)>-1) {
		    		/** session ID starts after the variable name (SESSION_ID_NAME) and "=" */ 
		    		int startPos = sessionID.indexOf(HTTPData.SESSION_ID_NAME) + HTTPData.SESSION_ID_NAME.length() + 1;
		    		int endPos;
		    		if (sessionID.indexOf(";",startPos)>-1) endPos = sessionID.indexOf(";",startPos);
		    		else endPos = sessionID.length();
		    		sessionID = sessionID.substring(startPos, endPos).trim();
		    	} else
		    		sessionID = generateSessionID(12);
		    	
				usaProxy.getProxyScript().send (client.getOut(), "proxyscript.js", client.getSocket(), sessionID);
			}
		    /** if data is to be logged */
			else if (logRequested) {
				usaProxy.getLogFile().log(client.getOut(), logData, client.getSocket());
			}
			/** all web server requests */
			else {
			    /** Open connection to the server */
			    try {	
			    	/** open a connection to the webserver */
			    	usaProxy.getProxyType().setRequestURL(new URL(get));
			    	this.server		= new SocketData(usaProxy.getProxyType().getServerConnect());
			    	
			    }
			    /** if server could not be detected or didn't respond */
			    catch (Exception err) {
		
					System.err.println("\nAn ERROR occured while connecting to the server:\n"
									+ err );
					err.printStackTrace();
		
					client.sendErrorMessage(err, get, host);
			
					/** Close streams and sockets */
					client.closeInputStream();
					client.closeOutputStream();
					client.closeSocket();
					if(server!=null) server.closeSocket();
			
					/** End of run */
					return;
					
			    }
			    /** Create server streams */
			    /** Create client streams */
			    server.getInputStream();
			    server.getOutputStream();
			    
			    /** serverdata ID is retrieved from HTTPTraffic.txt file */
			    if(usaProxy.getHttpTraffic().isLogHttpTraffic())
			    	httpTrafficIndex = usaProxy.getHttpTraffic().getHttpTrafficIndex(true);
			    
			    /** Create new Thread that manages the posting
			     *  of the client request to the web server */
			    new ServerRequest(client, server, this);
			    
			    /** Receive server response and send it to client */
				processResponse (server.getIn());
			}  
		}
		catch (Exception e) {
		    System.err.println("\nAn ERROR occured while negotiating with the server:\n"
		    				+ e);
		    e.printStackTrace();
		    
		}
		finally { 
		    
		    /** Close streams and sockets */
		    try { 
		    	client.closeInputStream();
				client.closeOutputStream();
				server.closeInputStream();
				server.closeOutputStream();
				client.closeSocket();
				server.closeSocket();
		    }
		    catch (Exception ex) { /**...*/ }
		}
    }
    
    /** Retrieves request headers and examines them in order to find out which action was requested.
     * <br><br>
     *  Modifies and adds headers (e.g. disables gzip encoding, X-Forwarded-For header).
     * 
     *  @param in is the client's InputStream.
     *  @return the remaining InputStream for further processing, or null in case.
     *  proxyscript.js was ordered or a log request was made.
     */
    private InputStream processRequest(InputStream in) {
    	
    	String method = "";
    	
    	try {
	        
    		in = client.readHeaders(in);
	        
	        /** retrieve url */
	        
	        String url	= (String) client.getHeaders().elementAt(0);	// get request-line
	        System.err.println(url);
	        if (!(HTTPData.isRequestLine(url))) {	// if first header != request-line search request-line
	        	
	        	for (int i=0; i<client.getHeaders().size(); i++) {
	        		if (HTTPData.isRequestLine(client.getHeaders().keyAt(i)))
	        			url = client.getHeaders().keyAt(i);
	        	}
	        }
			client.getHeaders().remove(url);
			method = url.substring(0,url.indexOf(" ")).trim();
			url = url.substring(url.indexOf(" ")).trim();		// leave url and http version
			
			/** delete HTTP version */
			if (url.toUpperCase().lastIndexOf("HTTP") >= 0) 	/** if HTTP version exists */
			    url = url.substring(0,url.toUpperCase().lastIndexOf("HTTP")).trim();
			else
			    url = url.trim();
			
			/** if testseite is ordered */
			if (url.toLowerCase().indexOf("usaproxylolo/testseite.html")!=-1) {
				testSeiteRequested = true;
				return null;
			}
			
			/** if js-file is ordered */
			if (url.toLowerCase().indexOf("usaproxylolo/proxyscript.js")!=-1) {
				proxyScriptRequested = true;
				return null;
			}
			
			/** if log request */
			if (url.toLowerCase().indexOf("usaproxylolo/img.jpg")!=-1) {
				logRequested = true;
				logData = url.substring(url.indexOf("?") + 1);
				return null;
			}
	
			/** Add local IP address in case a local address was requested */
			if(url.toLowerCase().startsWith("/"))
				url = "http://" + java.net.InetAddress.getLocalHost() + url;
			URL geturl = new URL(url);
			get = url;
			
			/** add new line with default HTTP/1.0 */
			client.getHeaders().add(0, usaProxy.getProxyType().getRequestLine(method, geturl),"");
	
			/** Display message: Client and requested URL */
			if (UsaProxy.DEBUG) System.out.println("New REQUEST:");
			if (UsaProxy.DEBUG) System.out.println("From client " + client.getSocket().getInetAddress().getHostName()
					+ " (" + client.getSocket().getInetAddress().getHostAddress() + ") with request GET " + url + "\n");
	        
	        /** retrieve host */
	        host = client.getHeaders().get(HTTPData.HEADER_HOST);
	        
	        /** add X-Forwarded-For header to ensure the server gets the client IP */
	        client.getHeaders().put(HTTPData.HEADER_X_FORWARDED_FOR, client.getSocket().getInetAddress().getHostAddress());
	        
	        /** add accept-encoding header value "identity" to avoid gzip content-encoding */
	        client.getHeaders().put(HTTPData.HEADER_ACCEPT_ENCODING, "identity");
	        
	        /** add connection header value "close" */
	        client.getHeaders().put(HTTPData.HEADER_CONNECTION, "close");
	        
	        /** remove proxy-connection and keep-alive header */
	        if(client.getHeaders().containsKey(HTTPData.HEADER_PROXY_CONNECTION))
	        	client.getHeaders().remove(HTTPData.HEADER_PROXY_CONNECTION);
	        if(client.getHeaders().containsKey(HTTPData.HEADER_CONNECTION_KEEPALIVE))
	        	client.getHeaders().remove(HTTPData.HEADER_CONNECTION_KEEPALIVE); 
	        
	        /** add pragma header value "no-cache" to avoid caching */
	        client.getHeaders().put(HTTPData.HEADER_PRAGMA, "no-cache");
	        
	        /** add UsaProxy header for Apache mod_rewrite */
	        client.getHeaders().put(HTTPData.HEADER_X_USAPROXY, "client");
        
        } catch (IOException e) {
        	e.printStackTrace();
        	if (client.getSocket()  != null) {
        		System.err.println("\nAn ERROR occured while processing request: "
        					+ client.getHeaders().elementAt(0) + " from client " + client.getSocket().getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while processing request: "
        					+ "\n"
							+ e );
        	}
		}
		if (method.equals("POST"))
			return in;
		else
			return null;
        
    }
    
    /** Retrieves response headers and data and examines them. If HTML was sent,
     *  response is enhanced with Javascript code (script element) and sent to the client.
     *  Else, the response is forwarded to the client unaltered.
     * <br><br>
     *  Modifies and adds headers (e.g. modifies content length, adds Via header).
     * 
     *  @param in is the server's InputStream.
     */
	private void processResponse (InputStream in) {
    	
    	/** Read the header from the server response */
		server.setHeaders(new OrderedHashtable());
        
        try {
        	in 	= server.readHeaders(in);
        
        } catch (IOException e) {
        	if (server.getSocket()  != null) {
        		System.err.println("\nAn ERROR occured while reading response headers: "
        					+ "from server " + server.getSocket().getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while reading response headers: "
        					+ "\n"
							+ e );
        	}
		}
        
        /** add VIA-header */
		try {
			java.net.InetAddress i = java.net.InetAddress.getLocalHost();
			String newVia = "";
	        if (server.getHeaders().containsKey(HTTPData.HEADER_VIA))
	        	newVia 	= (String) server.getHeaders().get(HTTPData.HEADER_VIA) + ", ";
	        newVia = newVia + i.getHostName() + " (UsaProxy/2.0)";
	        server.getHeaders().put(HTTPData.HEADER_VIA, newVia);
	    } catch (UnknownHostException e1) {
	    	System.err.println("\nAn ERROR occured while reading generating VIA header: " + e1 );
		}
        
        /** Check if html is being transmitted */
        String contentType = (String)server.getHeaders().get(HTTPData.HEADER_CONTENT_TYPE);
        if ((contentType != null) && HTTPData.isHTMLType(contentType))
        	isHtmlStream = true;
        else isHtmlStream = false;
        
        /** Check if 404 message is being transmitted
         *  if true no enhancing necessary*/
        if(UsaProxy.DEBUG) System.err.println("Status line: " + (String) server.getHeaders().elementAt(0) + "\n");
        String status	= (String) server.getHeaders().elementAt(0);	// get status-line
        if (!(HTTPData.isStatusLine(status))) {	// if first header != status-line search status-line
        	
        	for (int j=0; j<server.getHeaders().size(); j++) {
        		if (HTTPData.isStatusLine(server.getHeaders().keyAt(j)))
        			status = server.getHeaders().keyAt(j);
        			break;
        	}
        }
        if ((status!=null) && (status.indexOf("404") != -1))
        	isNotFoundMessage = true;
        else isNotFoundMessage = false;
        
        /** If html-file add ProxyScript */
        if (isHtmlStream && (isNotFoundMessage==false))
        	in = enhanceResponse(in);
        
        /** print all headers in OrderedHashtable server.getHeaders() */
        if (UsaProxy.DEBUG) { server.printHeaders(); }
        
        /** transmit response (header) */
        sendHeaders(client.getOut());
        sendResponse(); /** transmit either all data or, in case of html-data,
        				  * the rest after the proxyScript-part */
    }
    
	/** Parses the HTML code and pastes the ProxyScript script tag after the <head> tag.
     * 
     *  @param in is the server's InputStream.
     *  @return the remaining InputStream for further processing.
     */
    private InputStream enhanceResponse(InputStream in) {
    	
        byte b 					= 0;
        boolean readTooMuch 	= false;
        boolean noHead			= true;
        boolean headStarted		= false;
        this.data 				= new StringBuffer();
        byte[] bdata;
        String dataBefore;
        
        try {
        	
        	/** read in max 4097 bytes and search the head-tag */
	        for (ByteArrayOutputStream buf = new ByteArrayOutputStream(4097); ;) {
	            if (!readTooMuch) {
	                b = (byte) in.read();
	            }
	            if (b == -1) {
	                break;
	            }
	            readTooMuch = false;
	            if ((!headStarted && b != '<') || (headStarted && b != '>')) {
	                buf.write(b);
	            } else {    // b== '<' || (headStarted && b == '>')
	            	buf.write(b);
	            	
	            	if(!headStarted) {
		            	b = (byte) in.read();
		            	if ((b != 'h') && (b != 'H')) {
		            		buf.write(b);
		            		continue;
		            	}
		            	/** b== 'h' */
		            	buf.write(b);
		                b = (byte) in.read();
		                if ((b != 'e') && (b != 'E')) {
		                	buf.write(b);
		            		continue;
		            	}
		            	/** b== 'e' */
		            	buf.write(b);
		                b = (byte) in.read();
		                if ((b != 'a') && (b != 'A')) {
		                	buf.write(b);
		            		continue;
		            	}
		            	/** b== 'a' */
		            	buf.write(b);
		            	b = (byte) in.read();
		                if ((b != 'd') && (b != 'D')) {
		                	buf.write(b);
		            		continue;
		            	}
		            	/** b== 'd' */
		                headStarted = true;
		            	buf.write(b);
		            	b = (byte) in.read();
		                if (b != '>') {
		                	buf.write(b);
		            		continue;
		            	}
		            	buf.write(b);
	            	}
	                /** head-tag found */
	                System.out.println("\nUsaProxy script tag was added to request " + client.getHeaders().elementAt(0) + "\n"
	                				+ "Monitoring was started at client " + client.getSocket().getInetAddress().getHostName()
	                				+ " (" + client.getSocket().getInetAddress().getHostAddress() + ")");
	                headStarted = false;
	                readTooMuch = true;
	                noHead		= false;
	                buf.close();
	                bdata = buf.toByteArray();
	                buf.reset();
	                
	                dataBefore 		= new String(bdata, HTTPData.HEADER_DEFAULT_CHAR_ENCODING);
	                //String scriptString		= "<script src='http://lo.lo/usaproxylolo/proxyscript.js' type='text/javascript'></script>";
	                String scriptString		= usaProxy.getProxyType().getScriptString();
	                /** paste id attribute with the current serverdata ID */
	                scriptString			= insertID(scriptString);
	                
	                /** add processed data together with script string to the StringBuffer
	                 *  object that will be transmitted later */
	                data.append(dataBefore).append(scriptString);
	                
	                /** adapt header content-length */
	                if (server.getHeaders().containsKey(HTTPData.HEADER_CONTENT_LENGTH)) {
		                int contentLength = Integer.parseInt((String) server.getHeaders().get(HTTPData.HEADER_CONTENT_LENGTH));
		                contentLength += scriptString.getBytes().length;
		                server.getHeaders().put(HTTPData.HEADER_CONTENT_LENGTH, ""+contentLength);
	                }
	                break;
	            }
	            
	            /** if no head-tag was found simply add
		         *  processed data to the StringBuffer
		         *  that will be later transmitted */
		        if (noHead) {	
		        	buf.close();
		            bdata = buf.toByteArray();
		            buf.reset();
		            data.append(new String(bdata, HTTPData.HEADER_DEFAULT_CHAR_ENCODING));
		        }
        	}
        
        } catch (IOException e) {
        	if (server.getSocket() != null) {
        		System.err.println("\nAn ERROR occured while enhancing response: "
        					+ "from server " + server.getSocket().getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while enhancing response: "
        					+ "\n"
							+ e );
        	}
		}
        
		/** return the rest of the input stream */ 
        return in;
    }
    
    /** Adds an id attribute with the current serverdata (HTTPTrafficIndex) ID to the script tag
     * 
     *  @param scriptString is the script tag without id attribute.
     *  @return the script tag together with the current serverdata ID.
     */
    private String insertID(String scriptString) {
    	StringBuffer scriptStringID = new StringBuffer(scriptString);
        String proxyString = "<script ";
        int insertPos = scriptString.indexOf(proxyString) + proxyString.length();
        scriptStringID.insert(insertPos, "id='proxyScript_UsaProxy' ");
        scriptString = scriptStringID.toString();
        scriptStringID = new StringBuffer(scriptString);
        proxyString = "proxyscript.js";
        insertPos = scriptString.indexOf(proxyString) + proxyString.length();
        scriptStringID.insert(insertPos, "?id=" + httpTrafficIndex);
        scriptString = scriptStringID.toString();
		return scriptString;
	}

    /** Sends response headers to the client.
     *  <br><br>
     *  Logs a serverdata entry with the current HTTPTrafficIndex
     *  and stores headers in the corresponding txt-file.
     * 
     *  @param out is the client's OutputStream.
     */
	private void sendHeaders(OutputStream out) {
    	
		client.setOut(new DataOutputStream(new BufferedOutputStream(client.getOut())));
		StringBuffer headers = new StringBuffer();
		
		for (Iterator e = this.server.getHeaders().entrySet().iterator(); e.hasNext();) {
			
	        Entry me 	= (Entry) e.next();
	        String key 	= (String) me.getKey();
	        if (null == key) continue;
	        if (HTTPData.isStatusLine(key.trim()))
	        	headers.insert(0,key + "\r\n");
	        else {
	        	String value = (String)me.getValue();
	        	headers.append(key).append(": ").append(value).append("\r\n");
	        }
	    }
		headers.append("\r\n");
		
		if (UsaProxy.DEBUG) System.err.println("Enhanced response headers " + headers);
		
		try {
			client.getOut().write(headers.toString().getBytes(HTTPData.HEADER_DEFAULT_CHAR_ENCODING));
			client.getOut().flush();
			
			/** in case HTML is transmitted
			 *  create "serverdata" log entry with index already defined by serverRequest
			 *  as suffix for stored server data */
			if((isHtmlStream && (isNotFoundMessage==false)) && usaProxy.getHttpTraffic().isLogHttpTraffic()) {
				
				/** httpTraffic log entry */
				logData = dateStamp() + " " + get + " serverdata " + httpTrafficIndex;
				usaProxy.getLogFile().log(null, logData, client.getSocket());
				
				/** store headers to indexed txt-file */
				usaProxy.getHttpTraffic().store(httpTrafficIndex,"\r\n\r\n##### Response #####\r\n\r\n"+headers.toString());
			}
			
		} catch (IOException e) {
			if (client.getSocket()  != null) {
        		System.err.println("\nAn ERROR occured while sending response headers: "
        					+ "to client " + client.getSocket().getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while sending response headers: "
        					+ "\n"
							+ e );
        	}
		}	
    }
    
    
	/** Sends response to the client.
     *  <br><br>
     *  In case the response was enhanced before, the enhanced part is sent first
     *  followed by the rest of the response. Else, the complete response is simply forwarded.
     *  <br><br>
     *  The forwarded data is stored to the respective serverdata (HTTPTraffic) txt-file
     *  according to the current HTTPTrafficIndex.
     */
	private void sendResponse() {
    	
    	byte response[] = new byte[4096]; /** data buffer */  
        int responseLength;   /** length of response as number of bytes */
        
        try {
        	
        	/** transmit enhanced response data */
	        if (isHtmlStream && (isNotFoundMessage==false)) {
	        	client.getOut().write(data.toString().getBytes(HTTPData.HEADER_DEFAULT_CHAR_ENCODING));
	        	client.getOut().flush();
	        	
				/** store data to indexed txt-file */
				if(usaProxy.getHttpTraffic().isLogHttpTraffic()) usaProxy.getHttpTraffic().store(httpTrafficIndex, data.toString());	
	        }
        	
        	while ((responseLength = server.getIn().read(response)) != -1) {
				if (responseLength > 0) {
			/**		for(int gg=0; gg<response.length; gg++)
						System.out.print((char)response[gg]);  */
					client.getOut().write (response,0,responseLength);
					client.getOut().flush(); 
		            
		            /** in case HTML is transmitted
			 		  * store data to respective serverdata txt-file */
		            if((isHtmlStream && (isNotFoundMessage==false)) && usaProxy.getHttpTraffic().isLogHttpTraffic()) {
						/** store data to indexed txt-file */
		            	usaProxy.getHttpTraffic().store(httpTrafficIndex, response, 0, responseLength);
					} 
		        }
			}
		} catch (IOException e) {
			if (client.getSocket() != null) {
        		System.err.println("\nAn ERROR occured while sending response: "
        					+ "to client " + client.getSocket().getInetAddress().getHostName() + ":\n"
							+ e );
        	} else {
        		System.err.println("\nAn ERROR occured while sending response: "
        					+ "\n"
							+ e );
        	}
		}		
    }
    
    /** Returns String of the form "2004-12-31,23:59:59"
     *  
     *  @return the date stamp String
     *  */
	public String dateStamp() {
		Calendar now = Calendar.getInstance();
		return now.get(Calendar.YEAR) + "-" + completeDateVals(now.get(Calendar.MONTH)+1) + "-"
		  + completeDateVals(now.get(Calendar.DAY_OF_MONTH))
		  + "," + completeDateVals(now.get(Calendar.HOUR_OF_DAY))
		  + ":" + completeDateVals(now.get(Calendar.MINUTE))
		  + ":" + completeDateVals(now.get(Calendar.SECOND));
	}
	
	/** Converts one-digit numbers into two-digit ones with a leading zero
     *  
     *  @param dateVal is the digit to be extended
     *  @return the converted number
     *  */
	public String completeDateVals(int dateVal) {
		String dateVals = "" + dateVal;
		if (dateVals.length()<2) return "0" + dateVal;
		else return dateVals;
	}
	
	/** Returns the serverData (HTTPTraffic) class which manages the HTTPTraffic txt-file.
	 * 
	 *  @return the HTTPTraffic class
	 */
	public HTTPTraffic getHttpTraffic( ) {
		return usaProxy.getHttpTraffic();
	}

	/** Returns the serverData ID (httpTrafficIndex).
	 * 
	 *  @return the httpTrafficIndex.
	 */
	public int getHttpTrafficIndex() {
		return httpTrafficIndex;
	}

	/** Sets the serverDataIndex.
	 * 
	 *  @param httpTrafficIndex is the new httpTrafficIndex
	 */
	public void setHttpTrafficIndex(int httpTrafficIndex) {
		this.httpTrafficIndex = httpTrafficIndex;
	}
	
	/** Returns a positions-length random String.
	 * 
	 * @param positions is the length of the random String
	 * @return the random String
	 */
	public String generateSessionID(int positions) {

		int intI = 0;
		String strLargeText = "";
		int intValue = 0;
		String[] arrCharacters = { "0","1","2","3","4","5","6","7","8","9",
                               "A","B","C","D","E","F","G","H","I","J",
                               "K","L","M","N","O","P","Q","R","S","T",
                               "U","V","W","X","Y","Z","a","b","c","d",
                               "e","f","g","h","i","j","k","l","m","n",
                               "o","p","q","r","s","t","u","v","w","x","y","z" };
		/** Loop over number of characters in string. */
		for (intI = 0 ; intI < positions ; intI++){
			/** Get a random value between 0 and the length of the
			 * character list. */
			intValue = randRange_UsaProxy( 0, (arrCharacters.length - 1)); 
			/** Append a character that is randomly chosen */
			strLargeText += arrCharacters[ intValue ];
		}
		return strLargeText;
	}

	/** Returns a random index.
	 * 
	 *  @param intFrom is the lowest index to be selected
	 *  @param intTo is the highest index to be selected

	 *  @return the random index
	 */
	public int randRange_UsaProxy( double intFrom, double intTo){
		/** Make sure that we have rounded numbers. */
		intFrom = Math.floor(  intFrom );
		intTo = Math.floor(  intTo );
		 
		/** Return the random number. */
		return((int)(Math.floor(intFrom + ((intTo - intFrom + 1) * 
		/** Seed the random number if a value was passed. */
				Math.random()))));
	}

}