import java.util.Set;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

/** 
 *  UsaProxy - HTTP proxy for tracking and logging of user interactions on websites
    Copyright (C) 2006  Monika Wnuk - Media Informatics Group at the University of Munich
    <br><br>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    <br><br>
 *  Implements an Ordered Hashtable, with elements in
 *  chronological order (i.e. insertion order)
 *  with an additional index (Vector orderedKeysLower)
 *  for access to elements independent from lower/upper
 *  case notation of keys
 */

public class OrderedHashtable {

    private Vector    orderedKeys;
    private Vector    orderedKeysLower;
    private Hashtable hashTable;

    /**
     *  Constructor, creates an OrderedHashtable.
     */
    public OrderedHashtable() {
        orderedKeys 		= new Vector();
        orderedKeysLower 	= new Vector();
        hashTable 			= new Hashtable();
    }

    /**
     *  Constructor, creates an OrderedHashtable.
     *  @param initialCapacity is the initial size for the container.
     */
    public OrderedHashtable(int initialCapacity) {
        orderedKeys 		= new Vector(initialCapacity);
        orderedKeysLower 	= new Vector(initialCapacity);
        hashTable 			= new Hashtable(initialCapacity);
    }

    /**
     *  Maps the specified key to the specified value in this OrderedHashtable.
     *  The value can be retrieved by calling the get method with a key that is
     *  equal to the original key.
     *  @param key is the hashtable key.
     *  @param value is the value.
     *  @return the previous value of the specified key in this
     *  OrderedHashtable, or null if it did not have one.
     */
    @SuppressWarnings("unchecked")
	synchronized public String put(String key, String value) {
        int i = orderedKeys.indexOf(key);
        if (i == -1)  {
            //  Add new name/value pair.
            orderedKeys.addElement(key); // insert (append) to the end of the list
            orderedKeysLower.addElement(key.toLowerCase()); // insert (append) to the end of the list

        } else {
            //  Replace name/value pair.
            orderedKeys.setElementAt(key, i);
            orderedKeysLower.setElementAt(((String) key).toLowerCase(),i); // Replace name/value pair
        }
        return (String) hashTable.put(key, value);
    }
    
    /**
     *  Adds the specified String with the specified key to the OrderedHashtable.
     *  The key is added at the specified index to the orderedKeys and orderedKeysLower
     *  Vectors (in lower case to .orderedKeysLower)
     *  @param index is the position in orderedKeys and orderedKeysLower Vector
     *  the key is inserted at
     *  @param key is the hashtable key.
     *  @param value is the value.
     */
    @SuppressWarnings("unchecked")
    synchronized public void add(int index, String key, String value) {
    	int i = orderedKeys.indexOf(key);
        if (i == -1)  {
            //  Add new name/value pair.
            orderedKeys.add(index, key); 	// insert at the first position
            orderedKeysLower.add(index, key.toLowerCase()); // insert at the first position

        } else {
            //  Replace name/value pair.
            orderedKeys.setElementAt(key, i);
            orderedKeysLower.setElementAt(((String) key).toLowerCase(),i); // Replace name/value pair
        }
        hashTable.put(key, value);
    }

    
    /**
     *  Returns the value to which the specified key is mapped in this
     *  hashtable.
     *  @param lowerKey is a key in the OrderedHashtable.
     *  @return the value to which the key is mapped in this hashtable; null if
     *  the key is not mapped to any value in this hashtable.
     */
    synchronized public String get(String lowerKey) {
    	/** retrieve key from orderedKeys at index lower case key
    	 *  is stored at in orderedKeysLower */
    	if (getLowerIndex(lowerKey)>=0) {
    		String key = (String) orderedKeys.elementAt(getLowerIndex(lowerKey));
            return (String) hashTable.get(key);
    	}
    	else return null;
    }

    /**
     *  Returns an enumeration of the keys in this OrderedHashtable.
     *  @return an enumeration of the keys in this OrderedHashtable.
     */
    synchronized public Enumeration keys() {
        return orderedKeys.elements();
    }

    /**
     *  Returns an enumeration of the elements in this OrderedHashtable.
     *  @return an enumeration of the elements in this OrderedHashtable.
     */
    @SuppressWarnings("unchecked")
    synchronized public Enumeration elements() {
        int s = hashTable.size();
        Vector elements = new Vector(s);
        for (int i=0; i<s; i++) {
            elements.addElement(elementAt(i));
        }
        return elements.elements();
    }
    
    /**
     *  Returns a Set view of the entries contained in this OrderedHashtable.
     *  @return a Set of the entries in this OrderedHashtable.
     */
    synchronized public Set entrySet() {
        return hashTable.entrySet();
    }

    /**
     *  Returns the component at the specified index.
     *  @param index is an index into this OrderedHashtable.
     *  @return the <code>Object</code> component at the specified index.
     *  @throws ArrayIndexOutOfBoundsException if index is out of bounds.
     */
    synchronized public String elementAt(int index)
            throws ArrayIndexOutOfBoundsException {
    	if (index<orderedKeys.size()) {
    		String key = (String) orderedKeys.elementAt(index);
    		if ((String) hashTable.get(key) != "")
    			return (String) hashTable.get(key);
    		else return key;
    	} else {
    		return null;
    	}
    }

    /**
     *  Returns the key at the specified index.
     *  @param index is an index into this OrderedHashtable.
     *  @return the String key at the specified index.
     *  @throws ArrayIndexOutOfBoundsException if index is out of bounds.
     */
    synchronized public String keyAt(int index)
            throws ArrayIndexOutOfBoundsException {
    	if (index<orderedKeys.size()) {
    		return (String) orderedKeys.elementAt(index);
    	} else {
    		return null;
    	}
    }

    /**
     *  Returns the index of the specified <code>Object</code>.
     *  @param lowerKey is a key in the OrderedHashtable.
     *  @return the index of the specified <code>Object</code>.
     */
    synchronized public int getIndex(String lowerKey) {
    	/** retrieve key from orderedKeys at index lower case key
    	 *  is stored at in orderedKeysLower */
    	if (getLowerIndex(lowerKey)>=0) {
	    	String key = (String) orderedKeys.elementAt(getLowerIndex(lowerKey));
	        return orderedKeys.indexOf(key);
    	}
    	else return -1;
    }
    
    /**
     *  Returns the index of the specified String.
     *  @param key is a key in the OrderedHashtable.
     *  @return the index of the specified String.
     */
    synchronized public int getLowerIndex(String key) {
        return orderedKeysLower.indexOf(key.toLowerCase());
    }
    
    /**
     *  Tests if the specified String is a key in this OrderedHashtable.
     *  @param lowerKey is a key in the OrderedHashtable.
     *  @return if the specified String is a key in this OrderedHashtable
     */
    synchronized public boolean containsKey(String lowerKey) {
    	/** retrieve key from orderedKeys at index lower case key
    	 *  is stored at in orderedKeysLower */
    	if (getLowerIndex(lowerKey)>=0) {
    		String key = (String) orderedKeys.elementAt(getLowerIndex(lowerKey));
    		return hashTable.containsKey(key);
    	} else {
    		return false;
    	}
    }

    /**
     *  Removes the key (and its corresponding value) from this hashtable. This
     *  method does nothing if the key is not in the hashtable.
     *  @param lowerKey is the key that needs to be removed.
     */
    synchronized public void remove(String lowerKey) {
    	/** retrieve key from orderedKeys at index lower case key
    	 *  is stored at in orderedKeysLower */
    	String key = (String) orderedKeys.elementAt(getLowerIndex(lowerKey));
        orderedKeys.removeElement(key);
        hashTable.remove(key);
        orderedKeysLower.removeElement(lowerKey.toLowerCase());
    }

    /**
     * Removes an element at the specified index.
     * @param i is the index of the element to remove.
     */
    synchronized public void removeElementAt(int i) {
    	String key 		= (String) orderedKeys.elementAt(i);
    	orderedKeysLower.removeElementAt(i);
        orderedKeys.removeElementAt(i);
        hashTable.remove(key);
    }

    /**
     *  Clears this OrderedHashtable so that it contains no keys.
     */
    synchronized public void clear() {
    	orderedKeysLower.removeAllElements();
        orderedKeys.removeAllElements();
        hashTable.clear();
    }

    /**
     *  Returns the number of components in this OrderedHashtable.
     *  @return the number of components in this vector.
     */
    synchronized public int size() {
        return orderedKeys.size();
    }

    /**
     * Recomputes the OrderedHashtable capacity.
     * @param capacity is the capacity to ensure.
     */
    synchronized public void ensureCapacity(int capacity) {
    	orderedKeysLower.ensureCapacity(capacity);
        orderedKeys.ensureCapacity(capacity);
    }
}
