import java.io.*;
import java.net.*; 

/** 
 *  UsaProxy - HTTP proxy for tracking and logging of user interactions on websites
    Copyright (C) 2006  Monika Wnuk - Media Informatics Group at the University of Munich
    <br><br>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 *  <br><br>
 *  UsaProxy main class
 */
public class UsaProxy {
	
    private int 				port;			/**  UsaProxy port */
    private ProxyType			proxyType;		/**  type of UsaProxy module: true proxy, transparent proxy module
    											  *	 server module, or proxy within a network which is connected
    											  *  to the Internet over a gateway proxy (remote proxy) */
    private LogFile 			logFile;		/**  contains synchronized log()-method */
    private ProxyScript 		proxyScript;	/**  provides sendFile()-method */
    private HTTPTraffic 		httpTraffic;	/**  provides access to index and txt-file for storing
												  *  request and response data to */
    
    final static boolean 		DEBUG = false;	/** if true all debug-mode messages are printed
    
    /** Constructor: creates UsaProxy with port and UsaProxy type such as regular proxy,
     *  gateway proxy (remote proxy), transparent proxy, proxy module within web server
     *  @param port is the port the UsaProxy will run on
     *  @param proxyType represents the usage type of UsaProxy 
     */
	public UsaProxy(int port, ProxyType proxyType) {
		
		this.port 		= port;
		this.proxyType 	= proxyType;
		
		/** start UsaProxy-Server */
		proxyStart();
		
	}
    
	/** starts UsaProxy-Server (opens ServerSocket and waits for incoming connections) */
    private void proxyStart() {
    	
		/** Display message */
		System.out.println("Trying to start UsaProxy at port " + port);
		
		logFile 		= new LogFile();
		proxyScript 	= new ProxyScript();
		httpTraffic		= new HTTPTraffic();
	
		try {
	
		    /** Create server socket */
		    ServerSocket server = new ServerSocket(port); 
		    
		    /** Display start message */
			System.out.println("UsaProxy started at port " + port);
			System.out.println("UsaProxy ready for accepting incoming connections");
	
		    /** ProxyConnect endless loop */
		    while (true) { 
	
				/** waiting for client */
				Socket clientConnect = server.accept(); 
		
				/** Display message: new client with his address  */
				if (DEBUG) System.out.println("\nNew client connected:\n" + clientConnect.getInetAddress().getHostAddress());
		
				/** new ProxyRequest that will handle client requests */
				new ClientRequest (clientConnect, this);
		    } 

		/** Handle IO exception */
		} catch (IOException e) {
        	System.err.println("\nAn ERROR occured while starting UsaProxy:\n"
							+ e );
			System.exit(1);
		}
    }

    /** creates a new UsaProxy instance.
     *  <br><br>
     *  Depending on arguments the respective UsaProxy type is instantiated */
	public static void main(String[] args) {
		
		try {
			
			/** if gateway/remote proxy is specified 
			 *  In case UsaProxy is running on a computer within a LAN
			 *  or if proxy is specified as transparent/server proxy */
			if (args.length == 4) {	
				InetAddress remoteAddress = null;
	            try {	
	            	remoteAddress = InetAddress.getByName(args[1]);
	            }
	            catch(UnknownHostException e) {
	            	System.err.println("\nAn ERROR occured while retrieving remote proxy/server address:\n"
							+ e );
	                System.err.println("Usage: java UsaProxy <port> [<remote proxy/server ip-address>] [<remote proxy/server port>] [transparent|server]");
	                System.exit(1);
	            }
	            try {
		            if(Integer.parseInt(args[2]) >= 0 && Integer.parseInt(args[2]) <= 65535) {
		            	
		            	ProxyType proxyType;
		            	if(args[3].equals("transparent")) proxyType = new TransparentProxy(remoteAddress, Integer.parseInt(args[2]));
		            	else if(args[3].equals("transparentremote")) proxyType = new TransparentProxy(remoteAddress, Integer.parseInt(args[2]));
		            	else if(args[3].equals("server")) {
		            		proxyType = new ServerProxy(remoteAddress, Integer.parseInt(args[2]));
		            	}
		            	else proxyType = new RemoteProxy(remoteAddress, Integer.parseInt(args[2]));
			            /** Create UsaProxy with command line argument 1 as port number,
				    	 *  and type of proxy */
						new UsaProxy(Integer.parseInt(args[0]), proxyType);
		            }
		            else {
		            	System.err.println("\nAn ERROR occured while instantiating remote/transparent/server proxy!\n"
		            					+ "Usage: java UsaProxy <port> [<remote proxy/server ip-address>] "
		            					+ "[<remote proxy/server port>] [transparent|transparentremote|server|remote]");
		                System.exit(1);
		            }
	            }
	            catch (NumberFormatException e) {
	            	System.err.println("\nAn ERROR occured while binding UsaProxy to port "
	            					+ Integer.parseInt(args[0]) + " and remote port " + Integer.parseInt(args[2]) + ":\n"
									+ e );
	                System.exit(1);
	            }
		    }
			else if (args.length == 1) {		// regular proxy
				try {
					/** Create UsaProxy with command line argument as port number and regular proxy type */
					new UsaProxy(Integer.parseInt(args[0]), new Proxy());
				}
				catch (NumberFormatException e) {
					System.err.println("\nAn ERROR occured while binding UsaProxy to port "
	            					+ Integer.parseInt(args[0]) + ":\n"
									+ e );
	            	System.err.println("Usage: java UsaProxy <port>");
	                System.exit(1);
	            }
		    }
			else if (args.length == 0) {
				/** Default port */
				new UsaProxy(8000, new Proxy());
		    }
			else {
				System.err.println("\nAn ERROR occured while instantiating remote/transparent/server proxy!\n"
    					+ "False number of arguments!\n"
    					+ "Usage: java UsaProxy <port> [<remote proxy/server ip-address>]"
    					+ "[<remote proxy/server port>] [transparent|transparentremote|server|remote]");
				System.exit(1);
			}
		}
		catch (Exception e) {
		    System.err.println("\nAn ERROR occured while starting UsaProxy:\n"
									+ e );
		    System.exit(1);
		}
	}

	/**
	 * Returns the respective proxyType class which implements the ProxyType interface.
	 * @return the proxyType.
	 */
	public ProxyType getProxyType() {
		return proxyType;
	}

	/**
	 * Returns the logFile class.
	 * @return the logFile.
	 */
	public LogFile getLogFile() {
		return logFile;
	}

	/**
	 * Returns the proxyScript class.
	 * @return the proxyScript.
	 */
	public ProxyScript getProxyScript() {
		return proxyScript;
	}

	/**
	 * Returns the httpTraffic class.
	 * @return the httpTraffic.
	 */
	public HTTPTraffic getHttpTraffic() {
		return httpTraffic;
	}
	
	/**
	 * Returns the UsaProxy port number.
	 * @return the port.
	 */
	public int getPort() {
		return port;
	}
}
